#include "config.h"
#include <sys/types.h>
#include <netdb.h>
#include <string.h>
#include <syslog.h>
#include <memory.h>
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>
#include <grp.h>
#include <pwd.h>
#include <limits.h>

#include "misc.h"
#include "str.h"
#include "pset.h"
#include "sio.h"

#include "defs.h"  
#include "sconf.h"
#include "conf.h"
#include "attr.h"
#include "parse.h"
#include "addr.h"
#include "xconfig.h"

char *next_line() ;
void parsemsg() ;
void out_of_memory() ;
char **argv_alloc( int );

extern int line_count;
extern struct name_value socket_types[] ;
extern struct name_value service_types[] ;
extern struct name_value service_flags[];

char get_next_inet_entry( int fd, pset_h sconfs, struct service_config *defaults);

void parse_inet_conf_file( int fd, struct configuration *confp )
{
	pset_h sconfs                         = CNF_SERVICE_CONFS( confp );
	struct service_config *default_config = CNF_DEFAULTS( confp );
	struct service_config default_default_config;
	int ret;
	
	CLEAR( default_default_config );
	line_count = 0;

	for( ;; ) {
		
		ret = get_next_inet_entry(fd, sconfs, default_config);

		if( ret == -2 )
			break;
	}
}

char get_next_inet_entry( int fd, pset_h sconfs, struct service_config *defaults)
{
	char *p;
	str_h strp;
	char *line = next_line(fd);
	struct service_config *scp;
	unsigned u, i;
	char *func = "get_next_inet_entry";
	char *name = NULL, *rpcvers = NULL, *rpcproto = NULL;
	char *group, *proto;
	struct name_value *nvp;
	struct protoent *pep ;
	struct passwd *pw ;
	struct group *grp ;
	char *dot = ".";
	char *slash = "/";
	pset_h args;
	
	if( line == CHAR_NULL )
		return -2;

	strp = str_parse( line, " \t", STR_RETURN_ERROR, INT_NULL ) ;
	if( strp == NULL )
	{
		parsemsg( LOG_CRIT, func, "str_parse failed" ) ;
		return( -1 ) ;
	}

	if( (args = pset_create(10,10)) == NULL )
	{
		out_of_memory(func);
		return -1;
	}

	/* Break the line into components, based on spaces */
	while( (p = str_component( strp )) )
	{
		if( pset_add(args, p) == NULL )
		{
			parsemsg( LOG_CRIT, func, ES_NOMEM );
			return -1;
		}
	}

	/* get the service name */
	name = (char *)pset_pointer( args, 0 );

	/* Check to find the '/' for specifying RPC version numbers */
	if( (rpcvers = strstr(name, slash)) != NULL ) {
		*rpcvers = '\0';
		rpcvers++;
	}

	scp = sc_alloc( name );
	if( scp == NULL )
	{
		pset_destroy(args);
		return -1;
	}

	/* Get the socket type (stream dgram) */
	nvp = nv_find_value( socket_types, (char *)pset_pointer(args, 1) );
	if( nvp == NULL )
	{
		parsemsg( LOG_ERR, func, "Bad socket type: %s", p);
		pset_destroy(args);
		sc_free(scp);
		return -1;
	}

	scp->sc_socket_type = nvp->value;

	/* Get the protocol type */
	proto = (char *)pset_pointer(args,2);
	if( strstr(proto, "rpc") != NULL )
	{
		int rpcmin, rpcmax;
		struct rpc_data *rdp = SC_RPCDATA( scp ) ;

		if( rpcvers == NULL ) {
			pset_destroy(args);
			return -1;
			/* uh oh */
		}

		rpcmin = atoi(rpcvers);
		if( (rpcmin == LONG_MIN) || (rpcmin == LONG_MAX) ) {
			pset_destroy(args);
			return -1;
			/* Uh oh */
		}

		rpcmax = atoi(rpcvers+2);
		if( (rpcmax == LONG_MIN) || (rpcmax == LONG_MAX) ) {
			pset_destroy(args);
			return -1;
			/* Uh oh */
		}

		/* now have min and max rpc versions */
		rdp->rd_min_version = rpcmin;		
		rdp->rd_max_version = rpcmax;		

		rpcproto = strstr(proto, slash);
		*rpcproto = '\0';
		rpcproto++;
		proto = rpcproto;

		/* Set the RPC type field */
		nvp = nv_find_value( service_types, "RPC" );
		if ( nvp == NULL )
		{
			parsemsg( LOG_WARNING, func, "Bad foo %s", name ) ;
			pset_destroy(args);
			sc_free(scp);
			return -1;
		}

		M_SET(scp->sc_type, nvp->value);
	}
	if ( ( pep = getprotobyname( proto ) ) == NULL )
	{
		parsemsg( LOG_ERR, func, "Protocol %s not in /etc/protocols", proto ) ;
		pset_destroy(args);
		sc_free(scp);
		return -1;
	}

	scp->sc_protocol.name = make_string( 1, proto ) ;
	if ( scp->sc_protocol.name == NULL )
	{
		out_of_memory( func ) ;
		pset_destroy(args);
		sc_free(scp);
		return -1;
	}
	scp->sc_protocol.value = pep->p_proto;
	SC_SPECIFY(scp, A_PROTOCOL);

	/* Get the wait attribute */
	p = (char *)pset_pointer(args, 3);
	if ( EQ( p, "wait" ) )
		scp->sc_wait = YES ;
	else if ( EQ( p, "nowait" ) )
		scp->sc_wait = NO ;
	else
		parsemsg( LOG_ERR, func, "Bad value for wait: %s", p ) ;

	/* Get the user to run as */
	p = (char *)pset_pointer(args, 4);
	if( (group = strstr(p, dot)) )
	{
		*group = '\0';
		group++;
	
		grp = (struct group *)getgrnam( (char *)group ) ;
		if ( grp == NULL )
		{
			parsemsg( LOG_ERR, func, "Unknown group: %s", group ) ;
			pset_destroy(args);
			sc_free(scp);
			return -1;
		}	

		scp->sc_gid = ((struct group *)grp)->gr_gid;
		SC_SPECIFY( scp, A_GROUP );
	}

	pw = getpwnam( p );
	if ( pw == NULL )
	{
		parsemsg( LOG_ERR, func, "Unknown user: %s", p ) ;
		pset_destroy(args);
		sc_free(scp);
		return -1;
	}

	scp->sc_uid = pw->pw_uid;
	scp->sc_user_gid = pw->pw_gid;

	/* Get server name, or flag as internal */
	p = (char *)pset_pointer(args, 5);
	if( EQ( p, "internal" ) ) 
	{
		nvp = nv_find_value( service_types, "INTERNAL" );
		if ( nvp == NULL )
		{
			parsemsg( LOG_WARNING, func, "Bad foo %s", name ) ;
			pset_destroy(args);
			sc_free(scp);
			return -1;
		}

		M_SET(scp->sc_type, nvp->value);

		if( EQ( scp->sc_name, "time" ) ) {
			if( EQ( proto, "stream" ) )
				scp->sc_id = make_string(1, "time-stream");
			else
				scp->sc_id = make_string(1, "time-dgram");
		}

		if( EQ( scp->sc_name, "daytime" ) ) {
			if( EQ( proto, "stream" ) )
				scp->sc_id = make_string(1, "daytime-stream");
			else
				scp->sc_id = make_string(1, "daytime-dgram");
		}

		if( EQ( scp->sc_name, "chargen" ) ) {
			if( EQ( proto, "stream" ) )
				scp->sc_id = make_string(1, "chargen-stream");
			else
				scp->sc_id = make_string(1, "chargen-dgram");
		}

		if( EQ( scp->sc_name, "echo" ) ) {
			if( EQ( proto, "stream" ) )
				scp->sc_id = make_string(1, "echo-stream");
			else
				scp->sc_id = make_string(1, "echo-dgram");
		}

		if( EQ( scp->sc_name, "discard" ) ) 
		{
			parsemsg(LOG_WARNING, func, "service discard not supported");
			pset_destroy(args);
			sc_free(scp);
			return -1;
		}

			
	}else{

		scp->sc_server = make_string( 1, p );
		if ( scp->sc_server == NULL )
		{
			out_of_memory( func ) ;
			pset_destroy(args);
			sc_free(scp);
			return -1;
		}
		SC_SPECIFY( scp, A_SERVER);

		/* Get argv */ 
		scp->sc_server_argv = (char **)argv_alloc(pset_count(args)+1);

		for( u = 0; u < pset_count(args)-6 ; u++ )
		{
			p = make_string(1, (char *)pset_pointer(args, u+6));
			if( p == NULL )
			{
				for ( i = 1 ; i < u ; i++ )
					free( scp->sc_server_argv[i] );
				free( scp->sc_server_argv );
				pset_destroy(args);
				sc_free(scp);
				return -1;
			}
			scp->sc_server_argv[u] = p;
		}
		/* Set the reuse flag, as this is the default for inetd */
		nvp = nv_find_value( service_flags, "REUSE" );
		if ( nvp == NULL )
		{
			parsemsg( LOG_WARNING, func, "Bad foo %s", name ) ;
			pset_destroy(args);
			sc_free(scp);
			return -1;
		}
		M_SET(scp->sc_xflags, nvp->value);
	
		/* Set the NAMEINARGS flag, as that's the default for inetd */
		nvp = nv_find_value( service_flags, "NAMEINARGS" );
		if ( nvp == NULL )
		{
			parsemsg( LOG_WARNING, func, "Bad foo %s", name ) ;
			pset_destroy(args);
			sc_free(scp);
			return (-1);
		}
		M_SET(scp->sc_xflags, nvp->value);
		SC_SPECIFY( scp, A_SERVER_ARGS );

		if ( (scp->sc_id = make_string( 1, scp->sc_name )) )
			SC_PRESENT( scp, A_ID ) ;
		else
		{
			out_of_memory( func ) ;
			pset_destroy(args);
			sc_free(scp);
			return -1;
		}
	}
	
	SC_SPECIFY( scp, A_PROTOCOL );
	SC_SPECIFY( scp, A_USER );
	SC_SPECIFY( scp, A_SOCKET_TYPE );
	SC_SPECIFY( scp, A_WAIT );

	if( ! pset_add(sconfs, scp) )
	{
		out_of_memory( func );
		pset_destroy(args);
		sc_free(scp);
		return -1;
	}

	syslog( LOG_ERR, "added service %s", scp->sc_name);
	return 0;
}

